/*
 * JBoss, Home of Professional Open Source
 * Copyright 2007, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package sample.vpd.jdbc;

import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.Driver;
import java.sql.DriverPropertyInfo;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.SQLWarning;
import java.sql.Savepoint;
import java.sql.Statement;
import java.util.Map;
import java.util.Properties;

import javax.security.auth.Subject;
import javax.security.jacc.PolicyContext;
import javax.security.jacc.PolicyContextException;
import javax.servlet.http.HttpServletRequest;

import org.jboss.logging.Logger;

import oracle.jdbc.OracleDriver;
import oracle.jdbc.driver.OracleConnection;

/**
 * <p>Provides an implementation of the SQL Driver interface (java.sql.Driver) 
 * which can be wrapped around the
 * Oracle Java driver (oracle.jbbc.OracleDriver) and add an awareness of 
 * Oracle VPD functionality into a connection pool.</p>
 * 
 * <p>NOTE: this implementation is a proof-of-concept designed to show how
 * VPD awareness can be added to J2EE application without code modification
 * within the servlets/JSPs and EJBs that make up such an application. It 
 * has not been tested for robustness and should not be used in production
 * without further validation and development. More information and any 
 * available updates can be found at 
 * http://wiki.jboss.org/wiki/Wiki.jsp?page=IntegratingOracleVPDandJBoss.</p>
 * 
 * <p>Specifically, this class wraps OracleDriver so that, when OracleDriver
 * is asked for a Connection via the connect() method, what is actually
 * returned is an instance of ProxiedVPDConnection.</p>
 * 
 * <p>ProxiedVPDDriver is the class that injects the VPD proxy into 
 * the OracleConnection transparently - see this inner class for more
 * details</p>
 * 
 * <p>ProxiedVPDDriver is designed for use with JBoss 4.x and has been 
 * tested with JBoss4.2RC4. It works with Oracle 10g Enterprise Edition</p>
 * 
 * <p>ProxiedVPDDriver can be loaded in a JBoss server by adding a jar file with this
 * class in it to the $JBOSS_HOME/server/$PROFILE/deploy/lib directory (as 
 * well as the required underlying Oracle jar - ojdbc14.jar).</p>
 * 
 * <p>The XML file specifying the Oracle DataSource can then be modified so that 
 * instead of specifying oracle.jdbc.OracleDriver in the driver-class tag,
 * sample.vpd.jdbc.ProxiedVPDDriver is used. Note that the username and password
 * should be the proxy username and password details. Other fields remain unchanged.
 * </p>
 * 
 * @author <a href="pbennett@redhat.com">Pete Bennett</a>
 * @version $Revision: 1.1 $
 */
public final class ProxiedVPDDriver implements Driver
{
   private Logger log = Logger.getLogger(ProxiedVPDDriver.class);

   // The wrapped OracleDriver instance
   private OracleDriver oracleDriver;

   public ProxiedVPDDriver()
   {
      log.info("ProxiedVPDDriver Created");
   }

   /**
    * Creates a new instance of OracleDriver if required and stores it locally,
    * the local instance is then returned as required.
    * 
    * @return the current wrapped OracleDriver
    * @throws SQLException
    */
   private OracleDriver getOracleDriver() throws SQLException
   {
      if (oracleDriver == null)
      {
         log.info("OracleDriver delegate created");
         oracleDriver = new OracleDriver();
      }

      return oracleDriver;
   }

   /**
    * Intercepts calls which ask for a new Connection and delegates the call to the 
    * wrapped OracleDrive instance. The OracleConnection returned from OracleDriver
    * is then wrapped in a ProxiedVPDConnection and it is this class that is 
    * actually returned.
    */
   public Connection connect(String url, Properties props) throws SQLException
   {
      OracleConnection connection = (OracleConnection) getOracleDriver().connect(url, props);

      log.info("connect(url=" + url + ", properties=" + props);

      log.info("Connection class=" + connection.getClass().getName());

      return new ProxiedVPDConnection(connection);
   }

   // ALL METHODS BELOW SIMPLY DELEGATE TO THE WRAPPED ORACLE DRIVER

   // Call is delegated directly to wrapped OracleDriver
   public boolean acceptsURL(String url) throws SQLException
   {
      boolean retval = getOracleDriver().acceptsURL(url);

      log.info("acceptsURL(" + url + ")");

      return retval;
   }

   // Call is delegated directly to wrapped OracleDriver
   public int getMajorVersion()
   {
      return oracleDriver.getMajorVersion();
   }

   // Call is delegated directly to wrapped OracleDriver
   public int getMinorVersion()
   {
      return oracleDriver.getMinorVersion();
   }

   // Call is delegated directly to wrapped OracleDriver
   public DriverPropertyInfo[] getPropertyInfo(String name, Properties props) throws SQLException
   {
      return getOracleDriver().getPropertyInfo(name, props);
   }

   // Call is delegated directly to wrapped OracleDriver
   public boolean jdbcCompliant()
   {
      return oracleDriver.jdbcCompliant();
   }
   
   /**
    * <p>Provides an implementation of the SQL Connection interface (java.sql.Connection) 
    * which can be wrapped around the
    * Oracle implementation of the same (oracle.jbbc.OracleConnection) and add an awareness of 
    * Oracle VPD functionality into a connection instance.</p>
    * 
    * <p>NOTE: this implementation is a proof-of-concept designed to show how
    * VPD awareness can be added to J2EE application without code modification
    * within the servlets/JSPs and EJBs that make up such an application. It 
    * has not been tested for robustness and should not be used in production
    * without further validation and development. More information and any 
    * available updates can be found at 
    * http://wiki.jboss.org/wiki/Wiki.jsp?page=IntegratingOracleVPDandJBoss.</p>
    * 
    * <p>Specifically, this class wraps OracleConnection so that, when 
    * methods are called on this class, they are in fact executed 
    * on a ProxyConnection over the wrapped OracleConnection which has 
    * the username of the logged on user injected as the proxy user.</p>
    * 
    * <p>ProxiedVPDConnection is designed for use with JBoss 4.x and has been 
    * tested with JBoss4.2RC4. It works with Oracle 10g Enterprise Edition</p>
    * 
    * @author <a href="pbennett@redhat.com">Pete Bennett</a>
    * @version $Revision: 1.1 $
    */
   public class ProxiedVPDConnection implements Connection
   {
      // The wrapped OracleConnection delegate
      private OracleConnection connectionDelegate;

      // An instance of Properties to hold the proxy username
      private Properties proxyProperties = new Properties();

      // Create a new ProxiedVPDConnection that wraps the specified Oracle Connection
      ProxiedVPDConnection(OracleConnection delegate)
      {
         connectionDelegate = delegate;
      }

      /**
       * <p>Determines the authenticated user in the J2EE environment so that 
       * the username can be passed on to Oracle as the proxy username for a VPD 
       * aware database.</p>
       * 
       * <p>It uses the standard PolicyContext object and attempts to identify the currently
       * logged on user first from the Subject attribute and, failing this, from the 
       * HttpServletRequest attribute.</p>
       * 
       * @return a String representing the logged on username
       * @throws PolicyContextException
       */
      //TODO: this mechanism could be made more robust
      //TODO: this could be a configurable strategy
      private String getUsername() throws PolicyContextException
      {
         log.info("Asking for caller");

         Subject caller = (Subject) PolicyContext.getContext("javax.security.auth.Subject.container");
         HttpServletRequest httpServletRequest = (HttpServletRequest) PolicyContext
               .getContext("javax.servlet.http.HttpServletRequest");

         String username = null;

         if (caller != null && caller.getPrincipals() != null && caller.getPrincipals().iterator().hasNext())
         {
            username = caller.getPrincipals().iterator().next().getName();
         }
         else if (httpServletRequest != null && httpServletRequest.getUserPrincipal() != null)
         {
            username = httpServletRequest.getUserPrincipal().getName();
         }
         else
         {
            username = null;
         }

         return username;
      }
      
      /**
       * <p>Opens a proxy connection within the wrapped OracleConnection and sets the
       * proxy user name for the proxy connection to be equal to the user name
       * returned by getUsername().</p>
       * 
       * @return a Connection which is actually a proxy
       * @throws SQLException
       */
      private Connection openProxyConnection() throws SQLException
      {
         String username = null;

         try
         {
            username = getUsername();
         }
         catch (PolicyContextException ex)
         {
            ex.printStackTrace();
            throw new SQLException();
         }

         log.info("Setting proxy user name to " + username);

         //TODO: the proxy connection needs to be closed before it is opened again
         //TODO: it does not seem to cause an issue to close it before it has been opened for the first time - check this
         closeProxyConnection();
         
         // THE FOLLOWING TWO LINES OF CODE ESTABLISH A PROXY CONNECTION TO ORACLE VPD
         proxyProperties.setProperty(OracleConnection.PROXY_USER_NAME, username);
         connectionDelegate.openProxySession(OracleConnection.PROXYTYPE_USER_NAME, proxyProperties);

         return connectionDelegate;
      }
      
      /**
       * <p>Closes a proxy connection within the wrapped OracleConnection.</p>
       * 
       * @throws SQLException
       */
      private void closeProxyConnection() throws SQLException
      {
         connectionDelegate.close(OracleConnection.PROXY_SESSION);
      }

      // THE METHODS BELOW ARE ALL THOSE IN java.sql.Connection
      // THEY ARE IMPLEMENTED BY DELEGATING TO THE PROXY CONNECTION
      // RETURNED BY openProxyConnection()
      public void clearWarnings() throws SQLException
      {
         openProxyConnection().clearWarnings();
      }

      public void close() throws SQLException
      {
         openProxyConnection().close();
      }

      public void commit() throws SQLException
      {
         openProxyConnection().commit();
      }

      public Statement createStatement() throws SQLException
      {
         log.info("createStatement()");

         return openProxyConnection().createStatement();
      }

      public Statement createStatement(int resultSetType, int resultSetConcurrency) throws SQLException
      {
         return openProxyConnection().createStatement(resultSetType, resultSetConcurrency);
      }

      public Statement createStatement(int resultSetType, int resultSetConcurrency, int resultSetHoldability)
            throws SQLException
      {
         return openProxyConnection().createStatement(resultSetType, resultSetConcurrency, resultSetHoldability);
      }

      public boolean getAutoCommit() throws SQLException
      {
         return openProxyConnection().getAutoCommit();
      }

      public String getCatalog() throws SQLException
      {
         return openProxyConnection().getCatalog();
      }

      public int getHoldability() throws SQLException
      {
         return openProxyConnection().getHoldability();
      }

      public DatabaseMetaData getMetaData() throws SQLException
      {
         return openProxyConnection().getMetaData();
      }

      public int getTransactionIsolation() throws SQLException
      {
         return openProxyConnection().getTransactionIsolation();
      }

      public Map<String, Class<?>> getTypeMap() throws SQLException
      {
         return openProxyConnection().getTypeMap();
      }

      public SQLWarning getWarnings() throws SQLException
      {
         return openProxyConnection().getWarnings();
      }

      public boolean isClosed() throws SQLException
      {
         return openProxyConnection().isClosed();
      }

      public boolean isReadOnly() throws SQLException
      {
         return openProxyConnection().isReadOnly();
      }

      public String nativeSQL(String sql) throws SQLException
      {
         return openProxyConnection().nativeSQL(sql);
      }

      public CallableStatement prepareCall(String sql) throws SQLException
      {
         return openProxyConnection().prepareCall(sql);
      }

      public CallableStatement prepareCall(String sql, int resultSetType, int resultSetConcurrency) throws SQLException
      {
         return openProxyConnection().prepareCall(sql, resultSetType, resultSetConcurrency);

      }

      public CallableStatement prepareCall(String sql, int resultSetType, int resultSetConcurrency,
            int resultSetHoldability) throws SQLException
      {
         return openProxyConnection().prepareCall(sql, resultSetType, resultSetConcurrency, resultSetHoldability);

      }

      public PreparedStatement prepareStatement(String sql) throws SQLException
      {
         return openProxyConnection().prepareStatement(sql);
      }

      public PreparedStatement prepareStatement(String sql, int autoGeneratedKeys) throws SQLException
      {
         return openProxyConnection().prepareStatement(sql, autoGeneratedKeys);
      }

      public PreparedStatement prepareStatement(String sql, int[] columnIndexes) throws SQLException
      {
         return openProxyConnection().prepareStatement(sql, columnIndexes);
      }

      public PreparedStatement prepareStatement(String sql, String[] columnNames) throws SQLException
      {
         return openProxyConnection().prepareStatement(sql, columnNames);
      }

      public PreparedStatement prepareStatement(String sql, int resultSetType, int resultSetConcurrency)
            throws SQLException
      {
         return openProxyConnection().prepareStatement(sql, resultSetType, resultSetConcurrency);
      }

      public PreparedStatement prepareStatement(String sql, int resultSetType, int resultSetConcurrency,
            int resultSetHoldability) throws SQLException
      {
         return openProxyConnection().prepareStatement(sql, resultSetType, resultSetConcurrency, resultSetHoldability);
      }

      public void releaseSavepoint(Savepoint savepoint) throws SQLException
      {
         openProxyConnection().releaseSavepoint(savepoint);
      }

      public void rollback() throws SQLException
      {
         openProxyConnection().rollback();
      }

      public void rollback(Savepoint savepoint) throws SQLException
      {
         openProxyConnection().rollback(savepoint);
      }

      public void setAutoCommit(boolean autoCommit) throws SQLException
      {
         openProxyConnection().setAutoCommit(autoCommit);
      }

      public void setCatalog(String catalog) throws SQLException
      {
         openProxyConnection().setCatalog(catalog);
      }

      public void setHoldability(int holdability) throws SQLException
      {
         openProxyConnection().setHoldability(holdability);
      }

      public void setReadOnly(boolean readOnly) throws SQLException
      {
         openProxyConnection().setReadOnly(readOnly);
      }

      public Savepoint setSavepoint() throws SQLException
      {
         return openProxyConnection().setSavepoint();
      }

      public Savepoint setSavepoint(String name) throws SQLException
      {
         return openProxyConnection().setSavepoint(name);
      }

      public void setTransactionIsolation(int level) throws SQLException
      {
         openProxyConnection().setTransactionIsolation(level);
      }

      public void setTypeMap(Map<String, Class<?>> map) throws SQLException
      {
         openProxyConnection().setTypeMap(map);
      }
   }
}
